/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
   print.c

   Produces data suitable for printing on a Postscript device. I have made
   several assumptions to simplify things:
   - paper size is A4 = 595x842 points
   - margin on all sides = 72 points = 1 inch
   - 1 pixel = 1 point means no scaling is necessary
   - PS font has same geometry as X font
   - no background pattern
   - don't draw the grid
   - The lpr command is used for printing

   This works for me, but will probably break (more or less) for anyone
   not using my printer (a NEC S62P) with A4 paper.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Intrinsic.h>

#include "types.h"
#include "calc.h"
#include "../xstuff/fonts.h"

int paper_height = 842;
int paper_width = 595;
int margin = 72;
static int lastfont = 0;

extern int font_size[8];

static void 
set_font(FILE * fp, buffer * buf, int row, int col)
{
	int newfont = ret_font(buf->matrix, row, col);
	if (newfont == lastfont)
		return;
	lastfont = newfont;
	fprintf(fp, "/%s findfont\n", fonts[newfont].psfont);
	fprintf(fp, "%d scalefont\n", font_size[newfont & SIZE_MASK]);
	fprintf(fp, "setfont\n");
}

static void
print_cell(FILE * fp, XFontStruct * font_info, buffer * buf,
		int row, int col, int x_base, int y_base,
		unsigned int text_height)
{
	char b[80], *p;
	int x_pos = 5;
	int y_pos = (cell_height(buf, row) - text_height) / 2
	+ font_info->max_bounds.descent;
	int f = ret_format(buf->matrix, row, col);

	if (f & BORDER_MASK) {
		fprintf(fp, "newpath\n");
		if (f & BORDER_BOTTOM) {
			fprintf(fp, "%d %d moveto\n", x_base, y_base);
			fprintf(fp, "%d %d lineto\n",
				x_base + cell_width(buf, col), y_base);
		}
		if (f & BORDER_RIGHT) {
			if (!(f & BORDER_BOTTOM))
				fprintf(fp, "%d %d moveto\n",
				  x_base + cell_width(buf, col), y_base);
			fprintf(fp, "%d %d lineto\n",
				x_base + cell_width(buf, col),
				y_base + cell_height(buf, row));
		}
		if (f & BORDER_TOP) {
			if (!(f & BORDER_RIGHT))
				fprintf(fp, "%d %d moveto\n",
					x_base + cell_width(buf, col),
					y_base + cell_height(buf, row));
			fprintf(fp, "%d %d lineto\n",
				x_base, y_base + cell_height(buf, row));
		}
		if (f & BORDER_LEFT) {
			if (!(f & BORDER_TOP))
				fprintf(fp, "%d %d moveto\n",
				 x_base, y_base + cell_height(buf, row));
			fprintf(fp, "%d %d lineto\n",
				x_base, y_base);
		}
		fprintf(fp, "stroke\n");
	}
#if 0	/* don't draw the ordinary grid pattern */
	fprintf(fp, "newpath\n");
	fprintf(fp, "%d %d moveto\n", x_base, y_base);
	fprintf(fp, "%d %d lineto\n", x_base + cell_width(buf, col), y_base);
	fprintf(fp, "%d %d lineto\n", x_base + cell_width(buf, col),
		y_base + cell_height(buf, row));
	fprintf(fp, "%d %d lineto\n", x_base, y_base + cell_height(buf, row));
	fprintf(fp, "closepath\n");
	fprintf(fp, "stroke\n");
#endif

	b[0] = '\0';
	switch (ret_type(buf->matrix, row, col)) {
	case LABEL:
		set_font(fp, buf, row, col);
		sprintf(b, "%s", ret_taddr(buf->matrix, row, col));
		break;
	case CONSTANT:
	case EXPRESSION:
		set_font(fp, buf, row, col);
		sprintf(b, "%.2f", ret_val(buf->matrix, row, col));
		break;
	default:
		break;
	}

	fprintf(fp, "newpath\n");
	fprintf(fp, "%d %d moveto\n", x_base + x_pos, y_base + y_pos);

	fprintf(fp, "<");
	for (p = b; *p; p++)
		fprintf(fp, "%02x", *p);
	fprintf(fp, ">\n");

	fprintf(fp, "show\n");
}

static void plot_cell(FILE * fp, buffer * buf, int row, int col)
{
	switch (ret_type(buf->matrix, row, col)) {
	case CONSTANT:
	case EXPRESSION:
		fprintf(fp, "%g\n", ret_val(buf->matrix, row, col));
		break;
	default:
#ifdef DEBUG
		fprintf(stderr, "Contents of [%d,%d] is not a number",
			row, col);
#endif
		fprintf(fp, "\n");
		break;
	}
}

/* Check if the first row can be used for tics */
/* Yes, the heuristics are questionable */
static int is_tics(spread **matrix, int row, int c1, int c2)
{
	int c;

	if (c1 == c2)
		return (ret_type(matrix, row, c1) == LABEL);

	for (c = c1+1; c <= c2; c++)
		switch (ret_type(matrix, row, c)) {
		case LABEL:
			return TRUE;
		case CONSTANT:
		case EXPRESSION:
			return FALSE;
		default:
			break;
		}
	return FALSE;
}

/* Check if the contents of the first columns is likely to be titles */
static int is_titles(spread **matrix, int r1, int r2, int col)
{
	int r;

	if (r1 == r2)
		return (ret_type(matrix, r1, col) == LABEL);

	for (r = r1+1; r <= r2; r++)
		switch (ret_type(matrix, r, col)) {
		case LABEL:
			return TRUE;
		case CONSTANT:
		case EXPRESSION:
			return FALSE;
		default:
			break;
		}
	return FALSE;
}

void plot(int style)
{
	int has_tics, has_titles;
	int r0, r1, r2, r, c0, c1, c2, c;
	spread **matrix = w_list->buf->matrix;
	char fn_cmd[256], fn_data[256], fn_output[256], cmd[256];
	FILE *fp_cmd, *fp_data;
	int pid = getpid();

	sprintf(fn_cmd, "/tmp/siagplot%d.cmd", pid);
	sprintf(fn_output, "/tmp/siagplot%d.ps", pid);
	fp_cmd = fopen(fn_cmd, "w");
	fprintf(fp_cmd, "# This file is used by Siag to control Gnuplot\n");
	fprintf(fp_cmd, "set terminal postscript\n");
	fprintf(fp_cmd, "set output \"%s\"\n", fn_output);
	switch (style) {
	case POINTS:
		fprintf(fp_cmd, "set data style points\n");
		break;
	case LINESPOINTS:
		fprintf(fp_cmd, "set data style linespoints\n");
		break;
	case IMPULSES:
		fprintf(fp_cmd, "set data style impulses\n");
		break;
	case DOTS:
		fprintf(fp_cmd, "set data style dots\n");
		break;
	case STEPS:
		fprintf(fp_cmd, "set data style steps\n");
		break;
	case BOXES:
		fprintf(fp_cmd, "set data style boxes\n");
		break;
	case LINES:
	default:
		fprintf(fp_cmd, "set data style lines\n");
	}
	r0 = r1 = w_list->blku.row;
	r2 = w_list->blkl.row;
	c0 = c1 = w_list->blku.col;
	c2 = w_list->blkl.col;
	has_tics = is_tics(matrix, r0, c1, c2);
	has_titles = is_titles(matrix, r1, r2, c0);
	if (has_titles) c1++;
	if (has_tics) {
		char *p;

		r1++;
		fprintf(fp_cmd, "set xtics (");
		for (c = c1; c <= c2; c++) {
			if (c > c1)
				fprintf(fp_cmd, ", ");
			if ((p = ret_taddr(matrix, r0, c)))
				fprintf(fp_cmd, "\"%s\" ", p);
			fprintf(fp_cmd, "%d", c-c1);
		}
		fprintf(fp_cmd, ")\n");
	}
	fprintf(fp_cmd, "plot ");
	for (r = r1; r <= r2; r++) {
		char *p;

		sprintf(fn_data, "/tmp/siagplot%d.%d", pid, r);
		fp_data = fopen(fn_data, "w");
		fprintf(fp_cmd, "\"%s\"", fn_data);
		if (has_titles && ((p = ret_taddr(matrix, r, c0))))
			fprintf(fp_cmd, "title \"%s\"", p);
		if (r < r2)
			fprintf(fp_cmd, ", ");
		for (c = c1; c <= c2; c++) {
			plot_cell(fp_data, w_list->buf, r, c);
		}
		fclose(fp_data);
	}
	fprintf(fp_cmd, "\n");
	fclose(fp_cmd);
	sprintf(cmd, "gnuplot %s", fn_cmd);
	system(cmd);
	if (!fork()) {
		execlp("ghostview", "Ghostview",
		       "-landscape", fn_output, NULL);
		llpr("Can't exec ghostview");
#ifdef DEBUG
		fprintf(stderr, "Can't exec ghostview\n");
#endif
		exit(1);
	}
}

/* How to print the whole document without wasting paper

	The document can in theory contain up to 1000 rows by 1000 columns,
	but clearly in any real-world spreadsheet most of that area
	is unused and we would be printing hundreds of blank pages.
	The easy way is to check all the used rows, find the longest one
	and print the resulting square. This calls for a function print_area
	which takes the top left and bottom right coordinates as arguments
	and prints everything inbetween. The same function can
	also be used to print the block.
*/
static int page_height(buffer *buf, int r)
{
	int y_base, i;

	y_base = paper_height - margin;
	for (i = r; y_base > margin; i++)
		y_base -= cell_height(buf, i);
	return (i-r);
}

static int page_width(buffer *buf, int c)
{
	int x_base, j;

	x_base = margin;
	for (j = c; x_base+cell_width(buf, j) < paper_width - margin; j++)
		x_base += cell_width(buf, j);
	return (j-c);
}

static int
print_page(FILE *fp, buffer *buf, int fromr, int fromc, int tor, int toc)
{
	int i, j=0, x_base, y_base;
	unsigned int text_height;

	y_base = paper_height - margin;
	for (i = fromr; y_base > margin; i++) {
		y_base -= cell_height(w_list->buf, i);
		x_base = margin;
		for (j = fromc;
		     x_base + cell_width(w_list->buf, j) < paper_width - margin;
		     j++) {
			XFontStruct *font_info = font_struct(display,
				    ret_font(w_list->buf->matrix, i, j));
			text_height = font_height(display,
				    ret_font(w_list->buf->matrix, i, j));
			print_cell(fp, font_info, w_list->buf,
				   i, j, x_base, y_base, text_height);
			x_base += cell_width(w_list->buf, j);
		}
	}
	fprintf(fp, "showpage\n");
	return 0;
}

static int
print_area(buffer *buf, int fromr, int fromc, int tor, int toc)
{
	FILE *fp;
	char fn[256];
	int r, c;
	int pages;
	time_t t;

	sprintf(fn, "/tmp/siagprint%d.ps", getpid());
	fp = fopen(fn, "w");

	fprintf(fp, "%%!PS-Adobe-3.0 EPSF-3.0\n");
	fprintf(fp, "%%%%Creator: %s\n", version);
	fprintf(fp, "%%%%Title: Siag\n");	/* todo: print better title */
	t = time(NULL);
	fprintf(fp, "%%%%CreationDate: %s\n", ctime(&t));
	fprintf(fp, "%%%%Pages: (atend)\n");
	fprintf(fp, "%%%%PageOrder: Ascend\n");
	fprintf(fp, "%%%%BoundingBox: %d %d %d %d\n",
		0, 0, paper_width, paper_height);
	fprintf(fp, "%%%%Orientation: Portrait\n");
	fprintf(fp, "%%%%EndComments\n");

	pages = 0;
	for (r = fromr; r <= tor; r += page_height(buf, r)) {
		for (c = fromc; c <= toc; c += page_width(buf, c)) {
			pages++;
			lastfont = 0;	/* force setfont on top of page */
			fprintf(fp, "%%%%Page: %d %d\n", pages, pages);
			print_page(fp, buf, r, c, tor, toc);
		}
	}
	fprintf(fp, "%%%%Trailer:\n");
	fprintf(fp, "%%%%Pages: %d\n", pages);
	fprintf(fp, "%%%%EOF\n");
	fclose(fp);
	return 0;
}

void preview()
{
	char fn[256];
	int fromr, fromc, tor, toc;
	int i, clu;

	fromr = 1;
	fromc = 1;
	tor = line_last_used(w_list->buf->matrix);
	toc = 1;
	for (i = 1; i <= tor; i++) {
		clu = col_last_used(w_list->buf->matrix, i);
		if (clu > toc) toc = clu;
	}

	sprintf(fn, "/tmp/siagprint%d.ps", getpid());
	if (!print_area(w_list->buf, fromr, fromc, tor, toc)) {
		if (!fork()) {
			execlp("ghostview", "Ghostview", fn, NULL);
			llpr("Can't exec ghostview");
#ifdef DEBUG
			fprintf(stderr, "Can't exec ghostview\n");
#endif
			exit(1);
		}
	} else
		llpr("Can\'t make postscript\n");
}

void printer()
{
	char fn[256];
	int fromr, fromc, tor, toc;
	int i, clu;

	fromr = 1;
	fromc = 1;
	tor = line_last_used(w_list->buf->matrix);
	toc = 1;
	for (i = 1; i <= tor; i++) {
		clu = col_last_used(w_list->buf->matrix, i);
		if (clu > toc) toc = clu;
	}

	sprintf(fn, "/tmp/siagprint%d.ps", getpid());
	if (!print_area(w_list->buf, fromr, fromc, tor, toc)) {
		if (!fork()) {
			execlp("lpr", "lpr", fn, NULL);
			llpr("Can't exec lpr");
#ifdef DEBUG
			fprintf(stderr, "Can't exec lpr\n");
#endif
			exit(1);
		}
	}
}

